from google.cloud import bigquery

from google.cloud.exceptions import NotFound


def save_data_to_bq(insert_df, dataset_id, table_id):
    try:
        client = bigquery.Client()
        # 資料表參數
        table_ref = client.dataset(dataset_id).table(table_id)
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = True
        job = client.load_table_from_dataframe(insert_df, table_ref, job_config=job_config)
        job.result()
    except Exception as e:
        raise e


def query_multiple_data_from_bq(sql_list, data_name):
    try:
        client = bigquery.Client()
        result_dict = {}
        for sql, alias in zip(sql_list, data_name):
            results = client.query(sql)
            df = results.to_dataframe()
            result_dict[alias] = df
    except Exception as e:
        raise e
    return result_dict


def check_table_exist(table_id):
    client = bigquery.Client()

    try:
        client.get_table(table_id)  # Make an API request.
        print("Table {} already exists.".format(table_id))
    except NotFound:
        print("Table {} is not found.".format(table_id))


def create_table(table_name, columns, types, modes=None):
    client = bigquery.Client()

    if modes:
        schema = [bigquery.SchemaField(col, t, mode=mode) for col, t, mode in zip(columns, types, modes)]
    else:
        schema = [bigquery.SchemaField(col, t, mode="REQUIRED") for col, t in zip(columns, types)]
    table = bigquery.Table(table_name, schema=schema)
    table = client.create_table(table)  # Make an API request.
    print(
        "Created table {}.{}.{}".format(table.project, table.dataset_id, table.table_id)
    )



import re

from typing import NamedTuple, Dict, List


class TableInfo(NamedTuple):
    unique_keys: List
    date_time_col: str


def replace_table_names(sql, replacement, table_info_dict):
    from_pattern = r"(?i)\bFROM\s+([^\s()]+)"
    join_pattern = r"(?i)\bJOIN\s+([^\s()]+)"
    for parse_name, pattern in zip(["FROM", 'JOIN'], [from_pattern, join_pattern]):
        for i in re.finditer(pattern, sql):
            group = i.group()
            table_name = group.split(" ")[-1]
            only_table_name = table_name.split(".")[-1]
            table_info = table_info_dict[only_table_name]
            sql = re.sub(pattern,
                         r"{0} ({1})".format(parse_name,
                                             replacement.format(unique_keys=",".join(table_info.unique_keys),
                                                                date_time_col=table_info.date_time_col,
                                                                table=table_name)), sql)
    return sql


def generate_sql_for_bq(sql, table_info: Dict[str, TableInfo]):
    format_sql = """
    SELECT * FROM (
          SELECT
              *,
              ROW_NUMBER()
                  OVER (PARTITION BY {unique_keys})
                  row_number
          FROM {table}
          order by {date_time_col}
        )
        WHERE row_number = 1
    """
    result_sql = replace_table_names(sql, format_sql, table_info)
    return result_sql
